#include "./common.h"

#include <boost/filesystem.hpp>
#include <boost/program_options.hpp>
namespace po = boost::program_options;

typedef struct LogEntry {
    unsigned int timestamp;  // time > 0
    string name;             // person | batch file
    char type;               // DOCTOR | NURSE | BATCH
    char event;              // ARRIVAL | DEPARTURE
    int room;                // room # >= 0

    string token;  // authentication token
    string log;    // path to logile
} LogEntry;

LogEntry parse_options(int argc, char* argv[], bool batchOK) {
    // logappend -T <timestamp> -K <token> (-D <doctor-name> | -N <nurse-name>)
    //     (-A | -L) [-R <room-id>] <log>
    // logappend -B <file>

    po::options_description desc("Allowed options");
    desc.add_options()
        ("timestamp,T", po::value<int>(), "time the event is recorded")
        ("token,K", po::value<string>(), "token used to authenticate the log")
        ("doctor,D", po::value<string>(), "name of doctor")
        ("nurse,N", po::value<string>(), "name of nurse")
        ("arrive,A", "specify that current event is an arrival")
        ("leave,L", "specify that current event is a departure")
        ("room,R", po::value<int>(), "specifies the room ID for the event")
        ("batch,B", po::value<string>(), "specifies a batch file of commands")
        ("log,F", po::value<string>(),
            "path to the file containing the event log");

    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv).options(desc).run(), vm);
    } catch (...) {
        THROW "arg parse failed";
    }
    po::notify(vm);

    LogEntry e;

    // check for batch first due to different options
    if (vm.count("batch")) {
        if (!batchOK || vm.count("timestamp") || vm.count("token") ||
        vm.count("doctor") || vm.count("nurse") || vm.count("arrive") ||
        vm.count("leave") || vm.count("room") || vm.count("log")) {
            THROW "batch and non-batch";
        }
        e.timestamp = 0;
        e.name = "";
        e.type = BATCH;
        e.event = '\0';
        e.room = 0;
        e.token = "";
        e.log = vm["batch"].as<string>();    // path to batch file
        return e;
    }

    if (vm.count("timestamp")) {
        e.timestamp = vm["timestamp"].as<int>();
    } else {
        THROW "missing timestamp";
    }

    if (vm.count("token")) {
        e.token = vm["token"].as<string>();
    } else {
        THROW "missing token";
    }

    if (vm.count("doctor")) {
        e.name = vm["doctor"].as<string>();
        e.type = DOCTOR;
    } else if (vm.count("nurse")) {
        e.name = vm["nurse"].as<string>();
        e.type = NURSE;
    } else {
        THROW "missing type";
    }

    if (vm.count("arrive")) {
        e.event = ARRIVAL;
    } else if (vm.count("leave")) {
        e.event = DEPARTURE;
    } else {
        THROW "missing event";
    }

    if (vm.count("room")) {
        e.room = vm["room"].as<int>();
    } else {
        e.room = NOT_IN_ROOM;
    }

    if (vm.count("log")) {
        e.log = vm["log"].as<string>();
    } else {
        THROW "missing log";
    }

    return e;
}

bool validate_log_entry(LogEntry data, string log_data) {
    // process log data
    TheDataStructures ds = process_log(log_data);

    // validate log entry
    //   timestamp is after last timestamp
    if (data.timestamp <= ds.last_timestamp) {
        return false;
    }

    //   if entering the building,
    //     subject is not already in the building
    //   if entering a room,
    //     subject is in the building
    //     subject is not in any room
    //   if leaving a room,
    //     subject is in the building
    //     subject is in the room
    //   if leaving the building,
    //     subject is in the building
    //     subject is not in any room
    std::unordered_map<string, int>* map;
    if (data.type == DOCTOR) {
        map = &(ds.dr_location);
    } else {  // NURSE
        map = &(ds.nurse_location);
    }
    auto search = map->find(data.name);
    bool in_bldg = search != map->end();
    int curr_room = in_bldg ? search->second : NOT_IN_BLDG;

    bool entering_bldg = data.event == ARRIVAL && data.room == NOT_IN_ROOM;
    if (entering_bldg) return !in_bldg;

    bool entering_room = data.event == ARRIVAL && data.room >= 0;
    if (entering_room) return in_bldg && curr_room == NOT_IN_ROOM;

    bool leaving_room = data.event == DEPARTURE && data.room >= 0;
    if (leaving_room) return in_bldg && curr_room == data.room;

    bool leaving_bldg = data.event == DEPARTURE && data.room == NOT_IN_ROOM;
    if (leaving_bldg) return in_bldg && curr_room == NOT_IN_ROOM;

    return false;
}

unsigned int encrypt_and_hmac(string log_data, string token,
byte** new_file_data) {
    // encrypt
    //   new_ciphertext = encrypt(token, log_data)

    byte key[KEY_SIZE];
    derive_key(token, key);

    byte new_ciphertext[1<<20];
    unsigned int new_ciphertext_len;
    const byte* plaintext = reinterpret_cast<const byte*>(log_data.c_str());
    unsigned int plaintext_len = log_data.length();
    new_ciphertext_len = encrypt(plaintext, plaintext_len, key, new_ciphertext);

    // hmac
    unsigned int new_hmac = hmac(new_ciphertext, new_ciphertext_len, token);

    // log file format on disk: new_hmac | new_ciphertext
    unsigned int new_file_data_len = HMAC_SIZE + new_ciphertext_len;
    *new_file_data =
        reinterpret_cast<byte*>(malloc(new_file_data_len*sizeof(byte)));
    memcpy(*new_file_data, &new_hmac, HMAC_SIZE);
    memcpy(*new_file_data+HMAC_SIZE, new_ciphertext, new_ciphertext_len);

    return new_file_data_len;
}

void close_log(string log, string log_data, string token) {
    // encrypt and hmac log data, write to log file

    // encrypt and hmac
    byte* new_file_data;
    unsigned int new_file_data_len;
    new_file_data_len = encrypt_and_hmac(log_data, token, &new_file_data);

    // make subdirs if necessary
    size_t idx = log.find_last_of('/');
    if (idx != string::npos) {
        // found, get directory path substr
        string path = log.substr(0, idx);
        try {
            boost::filesystem::create_directories(path);
        } catch (...) {
            THROW ERROR_CONDITION;
        }
    }

    // write to disk
    ofstream out(log, std::ios::out | std::ios::binary);
    if (!out) {THROW ERROR_CONDITION;}
    out.write(reinterpret_cast<char*>(new_file_data), new_file_data_len);
    out.close();
}

void append(string &log_data, LogEntry data) {
    // validate log entry
    if (!validate_log_entry(data, log_data)) {THROW ERROR_CONDITION;}

    // append entry to log
    //   log_data += <record>
    char record[1<<10];
    sprintf(record, "%d,%c,%d,%c,%s", data.timestamp, data.event,
        data.room, data.type, data.name.c_str());
    log_data += record;
    log_data += "\n";
}

void do_batch(string batch_file) {
    ifstream command_lines(batch_file, std::ios::in);
    if (!command_lines) {THROW ERROR_CONDITION;}

    string line;
    vector<string> args;
    int argc;
    char** argv;
    LogEntry data;
    // map log files to log data
    // so files only have to get read in once and written once
    // better batch!
    std::unordered_map<string, string> log_data_map;
    std::unordered_map<string, string> token_map;
    while (getline(command_lines, line)) {
        args = split(line, " ");
        argc = args.size()+1;
        argv = reinterpret_cast<char**>(malloc(argc*sizeof(char*)));
        argv[0] = strdup("logappend");
        for (unsigned int i = 1; i <= args.size(); i++) {
            argv[i] = strdup(args[i-1].c_str());
        }
        try {
            data = parse_options(argc, argv, false);
            string log_data;
            auto search = log_data_map.find(data.log);
            if (search == log_data_map.end()) {
                // log file not yet opened
                // open it and save the data
                try {
                    log_data = open_log(data.log, data.token);
                } catch (const int err) {
                    if (err != LOG_NOT_OPEN) {
                        THROW err;
                    }
                }
                log_data_map[data.log] = log_data;
                token_map[data.log] = data.token;
            } else {
                // log file already opened
                log_data = search->second;
            }
            append(log_data, data);
            log_data_map[data.log] = log_data;
        } catch (const char*) {
            cout << "invalid" << endl;
        } catch (const int err) {
            if (err == ERROR_CONDITION) {
                cout << "invalid" << endl;
            } else if (err == INTEGRITY_VIOLATION) {
                cout << "integrity violation" << endl;
            } else {
                THROW err;
            }
        }
        args.clear();
    }
    // close all the logs
    for (auto iter : log_data_map) {
        string log = iter.first;
        close_log(log, iter.second, token_map[log]);
    }
}

int main(int argc, char* argv[]) {
    // read command line data
    LogEntry data;
    try {
        data = parse_options(argc, argv, true);
    } catch (const char*) {
        cout << "invalid" << endl;
        return ERROR_CONDITION;
    }

    try {
        if (data.type == BATCH) {
            do_batch(data.log);
        } else {  // single append
            string log_data;
            try {
                log_data = open_log(data.log, data.token);
            } catch (const int err) {
                if (err != LOG_NOT_OPEN) {
                    THROW err;
                }
            }
            append(log_data, data);
            close_log(data.log, log_data, data.token);
        }
    } catch (const int err) {
        if (err == ERROR_CONDITION)
            cout << "invalid" << endl;
        else if (err == INTEGRITY_VIOLATION)
            cout << "integrity violation" << endl;
        else
            std::cerr << "error code: " << err << endl;
        return ERROR_CONDITION;
    }

    return OK;
}
