#include <algorithm>

#include "./common.h"

#include <boost/program_options.hpp>
namespace po = boost::program_options;

#define STATUS 'S'
#define LIST_ROOMS 'R'

typedef struct CommandData {
    // common to STATUS and LIST_ROOMS
    string token;    // authentication token
    char action;    // STATUS | LIST_ROOM
    string log;        // path to logfile

    // these only for LIST_ROOMS
    string name;    // name of staff member
    char type;        // DOCTOR | NURSE
} CommandData;

CommandData parse_options(int argc, char* argv[]) {
    // logread -K <token> -S <log>
    // logread -K <token> -R (-D <doctor-name> | -N <nurse-name>) <log>

    // define command line options
    po::options_description desc("Allowed options");
    desc.add_options()
        ("token,K", po::value<string>(), "token used to authenticate the log")
        ("doctor,D", po::value<string>(), "name of doctor")
        ("nurse,N", po::value<string>(), "name of nurse")
        ("status,S", "print the current state of the log to stdout")
        ("list-rooms,R", "give a list of all rooms entered by a staff member")
        ("log,F", po::value<string>(),
            "path to the file containing the event log");

    // parse command line options
    po::variables_map vm;
    try {
        po::store(po::command_line_parser(argc, argv).options(desc).run(),
            vm);
    } catch (...) {
        THROW "arg parse failed";
    }
    po::notify(vm);

    CommandData data;

    // authentication token
    if (vm.count("token")) {
        data.token = vm["token"].as<string>();
    } else {
        THROW "missing token";
    }

    // query type
    if (vm.count("status")) {
        data.action = STATUS;
    } else if (vm.count("list-rooms")) {
        data.action = LIST_ROOMS;
    } else {
        THROW "missing action";
    }

    // log file
    if (vm.count("log")) {
        data.log = vm["log"].as<string>();
    } else {
        THROW "missing log";
    }

    // mandatory args depending on query type
    if (data.action == LIST_ROOMS) {
        if (vm.count("doctor")) {
            data.name = vm["doctor"].as<string>();
            data.type = DOCTOR;
        } else if (vm.count("nurse")) {
            data.name = vm["nurse"].as<string>();
            data.type = NURSE;
        } else {
            THROW "missing type";
        }
    }

    return data;
}

void print_join(vector<string> list, char delim) {
    // join a list of strings
    bool first = true;
    for (auto elem : list) {
        if (!first) {
            cout << delim;
        }
        cout << elem;
        first &= false;
     }
    cout << endl;
}

void print_staff(std::unordered_map<string, int> location) {
    // print list of staff

    vector<string> names;
    for (auto iter : location) {
        names.push_back(iter.first);
    }

    std::sort(names.begin(), names.end());
    print_join(names, ',');
}

void print_rooms(std::unordered_map<int, std::set<string>> room_occupants) {
    // print list of rooms and occupants

    // get list of occupied rooms
    int room_id;
    vector<int> rooms;
    for (auto iter : room_occupants) {
        room_id = iter.first;
        if (room_id >= 0 && !iter.second.empty()) {
            rooms.push_back(room_id);
        }
    }

    // print room-by-room occupant lists
    std::sort(rooms.begin(), rooms.end());
    for (auto r : rooms) {
        cout << r << ": ";
        vector<string> names;
        for (auto name : room_occupants[r]) {
            names.push_back(name);
        }
        std::sort(names.begin(), names.end());
        print_join(names, ',');
    }
}

/**
 * Print the current state of the log to stdout.
 * The state should be printed to stdout on at least two lines, with lines
 * separated by the \n (newline) character.
 * The first line should be a comma-separated list of doctors currently in the
 * hospital.
 * The second line should be a comma-separated list of nurses currently in the
 * hospital.
 * The remaining lines should provide room-by-room information indicating
 * which staff member is in which room.
 * Each line should begin with a room ID, printed as a decimal integer,
 * followed by a colon, followed by a space, followed by a comma-separated
 * list of staff members.
 * Room IDs should be printed in ascending integer order, all names should be
 * printed in ascending lexicographic string order.
 */
void status(const string log_data) {
    TheDataStructures data = process_log(log_data);
    print_staff(data.dr_location);
    print_staff(data.nurse_location);
    print_rooms(data.room_occupants);
}

/**
 * Give a list of all rooms entered by a staff member.
 * Output the list of rooms in chronological order.
 * The list is printed to stdout in one comma-separated list of room
 * identifiers.
 */
void list_rooms(const string log_data, const string target_name,
    const char target_type) {
    string log_data_copy = log_data;
    vector<string> lines = split(log_data_copy, "\n");

    // scan the log
    // print arrival events with room numbers
    bool first = true;
    for (string line : lines) {
        vector<string> elements = split(line, ",");
        char event = elements[1][0];
        int room = std::stoi(elements[2]);
        char type = elements[3][0];
        string name = elements[4];

        if (event == ARRIVAL && type == target_type && room >= 0 &&
            !name.compare(target_name)) {
            if (!first) {
                cout << ",";
            }
            cout << room;
            first &= false;
        }
    }
    cout << endl;
}

int main(int argc, char* argv[]) {
    CommandData data;
    try {
        data = parse_options(argc, argv);
    } catch (const char*) {
        cout << "invalid" << endl;
        return ERROR_CONDITION;
    }

    try {
        string log_data = open_log(data.log, data.token);
        if (data.action == STATUS) {
            status(log_data);
        } else {  // LIST_ROOMS
            list_rooms(log_data, data.name, data.type);
        }
    } catch (const int err) {
        if (err == ERROR_CONDITION || err == LOG_NOT_OPEN)
            cout << "invalid" << endl;
        else if (err == INTEGRITY_VIOLATION)
            cout << "integrity violation" << endl;
        else
            std::cerr << "error code: " << err << endl;
        return ERROR_CONDITION;
    }

    return OK;
}
